/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/time.h>
#include <jeffpc/val.h>
#include <jeffpc/sexpr.h>
#include <jeffpc/io.h>
#include <jeffpc/unicode.h>

#include <ctype.h>
#include <locale.h>
#include <ncurses.h>
#include <math.h>

#include "version.h"
#include "qso.h"
#include "adif.h"
#include "country.h"
#include "maidenhead.h"

#define EARTH_RADIUS		6378.137

#define ENTRY_WIN_HEIGHT	(6 + 2)
#define ENTRY_WIN_WIDTH		(80 + 2)
#define ENTRY_WIN_ROW		0
#define CALL_WIN_HEIGHT		11
#define CALL_WIN_WIDTH		ENTRY_WIN_WIDTH
#define CALL_WIN_ROW		(ENTRY_WIN_ROW + ENTRY_WIN_HEIGHT)
#define CMD_WIN_HEIGHT		3
#define CMD_WIN_WIDTH		RECENT_WIN_WIDTH
#define CMD_WIN_ROW		(LINES - 3)
#define RECENT_WIN_HEIGHT	(LINES - ENTRY_WIN_HEIGHT - CALL_WIN_HEIGHT - \
				 CMD_WIN_HEIGHT)
#define RECENT_WIN_WIDTH	CALL_WIN_WIDTH
#define RECENT_WIN_ROW		(CALL_WIN_ROW + CALL_WIN_HEIGHT)

#define CMD_HINT_COLOR		1

static const char *data_dir;
static int data_dir_fd;

static struct str default_date = STR_STATIC_INITIALIZER("????""-??""-??");
static struct str default_time = STR_STATIC_INITIALIZER("??:??:??");
static struct str default_info =
	STR_STATIC_INITIALIZER("??                                 ");
static struct str default_latlon = STR_STATIC_INITIALIZER("??.??N ???.??E");
static struct str default_azidist =
	STR_STATIC_INITIALIZER("???° ????? km/???° ????? km");

/* This contains everything being displayed on screen */
static struct state {
	struct qso qso;

	/* the following are generated from the QSO data above */
	struct str *date;
	struct str *time_start;
	struct str *time_end;
	struct str *info;
	struct str *latlon; /* lat & lon */
	struct str *azidist; /* azimuth & distance */
} state = {
	.date = &default_date,
	.time_start = &default_time,
	.time_end = &default_time,
	.info = &default_info,
	.latlon = &default_latlon,
	.azidist = &default_azidist,
};

enum refresh {
	REF_DE =		0x00000001,
	REF_DE_GRID =		0x00000002,
	REF_DE_POWER =		0x00000004,
	REF_TIMES =		0x00000008,
	REF_CALL =		0x00000010,
	REF_FREQ =		0x00000020,
	REF_MODE =		0x00000040,
	REF_RSTS =		0x00000080,
	REF_RSTR =		0x00000100,
	REF_NAME =		0x00000200,
	REF_QTH =		0x00000400,
	REF_GRID =		0x00000800,
	REF_INFO =		0x00001000,
	REF_GEO =		0x00002000,
	REF_STATE =		0x00004000,
	REF_COUNTY =		0x00008000,
	REF_ZONE_CONT =		0x00010000,
	REF_ZONE_DXCC =		0x00020000,
	REF_ZONE_ITU =		0x00040000,
	REF_ZONE_WAZ =		0x00080000,
#define REF_ZONES	(REF_ZONE_CONT | REF_ZONE_DXCC | REF_ZONE_ITU | \
			 REF_ZONE_WAZ)
	REF_COMMENT =		0x00100000,
};

struct entry_field {
	int row;
	int col_label;
	int col_value;
	/*
	 * The field label.  All characters other than '_' are printed.  A
	 * '_' indicates that the following character should be highlighted
	 * to hint at the command to modify this field.
	 */
	const char *label;
	size_t len;
	enum val_type type;
	enum refresh refresh;
	ptrdiff_t offset;
};

struct command {
	const char *cmd;
	void (*fxn)(char *arg);
	bool arg_required;
	enum refresh refresh;
};

static void cmd_new(char *arg);
static void cmd_reset(char *arg);
static void cmd_save(char *arg);
static void cmd_start(char *arg);
static void cmd_end(char *arg);
static void cmd_de(char *arg);
static void cmd_de_grid(char *arg);
static void cmd_de_power(char *arg);
static void cmd_call(char *arg);
static void cmd_freq(char *arg);
static void cmd_mode(char *arg);
static void cmd_rst_s(char *arg);
static void cmd_rst_r(char *arg);
static void cmd_name(char *arg);
static void cmd_qth(char *arg);
static void cmd_grid(char *arg);
static void cmd_state(char *arg);
static void cmd_county(char *arg);
static void cmd_continent(char *arg);
static void cmd_dxcc(char *arg);
static void cmd_itu(char *arg);
static void cmd_waz(char *arg);
static void cmd_comment(char *arg);

static const struct entry_field entry_fields[] = {
	{ 0,  0,  3, "_d_e",          10, VT_STR, REF_DE,
	  offsetof(struct state, qso.me.call), },
	{ 0, 15, 21, "_Grid:",         8, VT_STR, REF_DE_GRID,
	  offsetof(struct state, qso.me.grid), },
	{ 0, 31, 36, "_P_w_r:",        5, VT_INT, REF_DE_POWER,
	  offsetof(struct state, qso.me.power), },
	{ 0,  0, 52, NULL,            10, VT_STR, REF_TIMES,
	  offsetof(struct state, date), },
	{ 0,  0, 63, NULL,             8, VT_STR, REF_TIMES,
	  offsetof(struct state, time_start), },
	{ 0,  0, 72, NULL,             8, VT_STR, REF_TIMES,
	  offsetof(struct state, time_end), },
	{ 1,  0,  6, "_Call:",        10, VT_STR, REF_CALL,
	  offsetof(struct state, qso.call), },
	{ 1, 18, 24, "_Freq:",        10, VT_INT, REF_FREQ,
	  offsetof(struct state, qso.freq), },
	{ 1, 36, 42, "_Mode:",         5, VT_STR, REF_MODE,
	  offsetof(struct state, qso.mode), },
	{ 1, 49, 55, "_RST_s:",        3, VT_STR, REF_RSTS,
	  offsetof(struct state, qso.rsts), },
	{ 1, 60, 66, "_RST_r:",        3, VT_STR, REF_RSTR,
	  offsetof(struct state, qso.rstr), },
	{ 2,  0,  6, "_Na_me:",       16, VT_STR, REF_NAME,
	  offsetof(struct state, qso.name), },
	{ 2, 24, 29, "_Q_T_H:",       35, VT_STR, REF_QTH,
	  offsetof(struct state, qso.qth), },
	{ 2, 66, 72, "_Grid:",         8, VT_STR, REF_GRID,
	  offsetof(struct state, qso.grid), },
	{ 3,  0,  0, NULL,            35, VT_STR, REF_INFO,
	  offsetof(struct state, info), },
	{ 3,  0, 37, NULL,            14, VT_STR, REF_GEO,
	  offsetof(struct state, latlon), },
	{ 3,  0, 53, NULL,            27, VT_STR, REF_GEO,
	  offsetof(struct state, azidist), },
	{ 4,  0,  7, "_S_t/Pr:",       2, VT_STR, REF_STATE,
	  offsetof(struct state, qso.state), },
	{ 4, 11, 19, "_C_o_u_n_ty:",  22, VT_STR, REF_COUNTY,
	  offsetof(struct state, qso.county), },
	{ 4, 43, 49, "_C_o_n_t:",      2, VT_STR, REF_ZONE_CONT,
	  offsetof(struct state, qso.continent), },
	{ 4, 53, 59, "_D_X_C_C:",      3, VT_STR, REF_ZONE_DXCC,
	  offsetof(struct state, qso.dxcc), },
	{ 4, 64, 69, "_I_T_U:",        2, VT_INT, REF_ZONE_ITU,
	  offsetof(struct state, qso.itu), },
	{ 4, 73, 78, "_W_A_Z:",        2, VT_INT, REF_ZONE_WAZ,
	  offsetof(struct state, qso.waz), },
	{ 5,  0,  9, "_Co_mment:",    71, VT_STR, REF_COMMENT,
	  offsetof(struct state, qso.comment), },
};

static const struct command commands[] = {
	{ "new",    cmd_new,		false, ~0u },
	{ "reset",  cmd_reset,		false, ~0u },
	{ "save",   cmd_save,		false, 0 },
	{ "start",  cmd_start,		true,  REF_TIMES, },
	{ "end",    cmd_end,		true,  REF_TIMES, },

	{ "de",     cmd_de,		true,  REF_DE, },
	{ "dg",     cmd_de_grid,	true,  REF_DE_GRID | REF_GEO, },
	{ "dpwr",   cmd_de_power,	true,  REF_DE_POWER, },
	{ "c",      cmd_call,		true,  REF_CALL | REF_INFO | REF_GEO |
					       REF_ZONES, },
	{ "f",      cmd_freq,		true,  REF_FREQ, },
	{ "m",      cmd_mode,		true,  REF_MODE, },
	{ "rs",     cmd_rst_s,		true,  REF_RSTS, },
	{ "rr",     cmd_rst_r,		true,  REF_RSTR, },
	{ "nm",     cmd_name,		true,  REF_NAME, },
	{ "qth",    cmd_qth,		true,  REF_QTH, },
	{ "g",      cmd_grid,		true,  REF_GRID | REF_GEO, },
	{ "st",     cmd_state,		true,  REF_STATE, },
	{ "count",  cmd_county,		true,  REF_COUNTY, },
	{ "cont",   cmd_continent,	true,  REF_ZONE_CONT, },
	{ "dxcc",   cmd_dxcc,		true,  REF_ZONE_DXCC, },
	{ "itu",    cmd_itu,		true,  REF_ZONE_ITU, },
	{ "waz",    cmd_waz,		true,  REF_ZONE_WAZ, },
	{ "cm",     cmd_comment,	true,  REF_COMMENT, },
};

#define TIME_ROW	0
#define TIME_COL	(80 - 28)

static WINDOW *entry_win;
static WINDOW *call_win;
static WINDOW *recent_win;
static WINDOW *cmd_win;

/*
 * Command handler helpers
 */
static void __recalculate_time(const struct time *time, struct str **out)
{
	if (time->have & TIME_PART_TIME_HM) {
		char buf[9];

		if (time->have & TIME_PART_TIME_S)
			snprintf(buf, sizeof(buf), "%02u:%02u:%02u",
				 time->hour, time->min, time->sec);
		else
			snprintf(buf, sizeof(buf), "%02u:%02u:__",
				 time->hour, time->min);

		*out = STR_DUP(buf);
	} else {
		*out = &default_time;
	}
}

static void recalculate_times(void)
{
	struct time *date;
	char buf[11];

	str_putref(state.date);
	str_putref(state.time_start);
	str_putref(state.time_end);

	if (state.qso.start.have & TIME_PART_DATE)
		date = &state.qso.start;
	else if (state.qso.end.have & TIME_PART_DATE)
		date = &state.qso.end;
	else
		date = NULL;

	if (!date) {
		state.date = &default_date;
	} else {
		snprintf(buf, sizeof(buf), "%04u-%02u-%02u", date->year,
			 date->month, date->day);

		state.date = STR_DUP(buf);
	}

	__recalculate_time(&state.qso.start, &state.time_start);
	__recalculate_time(&state.qso.end, &state.time_end);
}

static void recalculate_location(void)
{
	double lat, lon;

	str_putref(state.latlon);

	if (maidenhead_decode(str_cstr(state.qso.grid), &lat, &lon) ||
	    country_latlon(state.qso.call, &lat, &lon)) {
		char buf[15];
		char ns, ew;

		ns = (lat >= 0) ? 'N' : 'S';
		ew = (lon >= 0) ? 'E' : 'W';

		snprintf(buf, sizeof(buf), "%5.2f%c %6.2f%c",
			 fabs(lat), ns,
			 fabs(lon), ew);

		state.latlon = STR_DUP(buf);
	} else {
		state.latlon = &default_latlon;
	}
}

static void recalculate_distance(void)
{
	double de_lat, de_lon;
	double dx_lat, dx_lon;
	double distance;
	double azimuth;
	unsigned int int_azi;
	char buf[30];

	str_putref(state.azidist);

	if (!state.qso.me.grid)
		goto not_calculated; /* no idea where we are */
	if (!state.qso.call && !state.qso.grid)
		goto not_calculated; /* no idea where they are */

	if (!maidenhead_decode(str_cstr(state.qso.me.grid), &de_lat, &de_lon))
		goto not_calculated; /* no idea where we are */

	if (!maidenhead_decode(str_cstr(state.qso.grid), &dx_lat, &dx_lon) &&
	    !country_latlon(state.qso.call, &dx_lat, &dx_lon))
		goto not_calculated; /* no idea where they are */

	/* degrees to radians */
	de_lat *= M_PI / 180.;
	de_lon *= M_PI / 180.;
	dx_lat *= M_PI / 180.;
	dx_lon *= M_PI / 180.;

	/*
	 * The following assumes a spherical earth.  This makes all the
	 * numbers slightly off from what we'd get if we used WGS84
	 * math...which we should.
	 */

	/* calculate */
	distance = acos(sin(de_lat) * sin(dx_lat) +
			cos(de_lat) * cos(dx_lat) * cos(de_lon - dx_lon));
	azimuth = fmod(atan2(sin(dx_lon - de_lon) * cos(dx_lat),
			     cos(de_lat) * sin(dx_lat) -
			     sin(de_lat) * cos(dx_lat) * cos(dx_lon - de_lon)),
		       2 * M_PI);

	/* radians to km & degrees */
	distance *= (2 * M_PI * EARTH_RADIUS) / (2 * M_PI);
	azimuth *= 180 / M_PI;

	/* convert to an int */
	int_azi = (azimuth >= 0.) ? azimuth : (azimuth + 360.);

	snprintf(buf, sizeof(buf), "%3u° %5.0f km/%3u° %5.0f km",
		 int_azi, distance,
		 (int_azi < 180) ? (int_azi + 180) : (int_azi - 180),
		 2 * M_PI * EARTH_RADIUS - distance);

	state.azidist = STR_DUP(buf);
	return;

not_calculated:
	state.azidist = &default_azidist;
}

/*
 * Command handlers
 */
static void cmd_new(char *arg)
{
	int ret;

	ret = qso_save(&state.qso, data_dir_fd);
	if (ret) {
		wprintw(recent_win, "Failed to save QSO: %s",
			xstrerror(ret));
		return;
	}

	qso_reset(&state.qso, false);

	str_putref(state.info);
	state.info = &default_info;

	recalculate_times();
	recalculate_location();
	recalculate_distance();
}

static void cmd_reset(char *arg)
{
	qso_reset(&state.qso, false);

	str_putref(state.info);
	state.info = &default_info;

	recalculate_times();
	recalculate_location();
	recalculate_distance();
}

static void cmd_save(char *arg)
{
	int ret;

	ret = qso_save(&state.qso, data_dir_fd);
	if (ret) {
		wprintw(recent_win, "Failed to save QSO: %s",
			xstrerror(ret));
		return;
	}
}

static void cmd_start(char *arg)
{
	if (!strcasecmp(arg, "now"))
		qso_set_datetime(&state.qso, true, NULL);
	else
		qso_set_datetime(&state.qso, true, arg);

	recalculate_times();
}

static void cmd_end(char *arg)
{
	if (!strcmp(arg, "now"))
		qso_set_datetime(&state.qso, false, NULL);
	else
		qso_set_datetime(&state.qso, false, arg);

	recalculate_times();
}

static void cmd_de(char *arg)
{
	qso_set_call(&state.qso, true, arg);
}

static void cmd_de_grid(char *arg)
{
	qso_set_grid(&state.qso, true, arg);

	recalculate_distance();
}

static void cmd_de_power(char *arg)
{
	qso_set_power(&state.qso, arg);
}

static void cmd_call(char *arg)
{
	struct country_result result;

	qso_set_call(&state.qso, false, arg);

	str_putref(state.info);
	str_putref(state.qso.continent);
	str_putref(state.qso.dxcc);

	if (country_lookup(state.qso.call, &result)) {
		state.info = result.name;
		state.qso.continent = result.continent;
		state.qso.dxcc = result.prefix;
		state.qso.itu = result.itu;
		state.qso.waz = result.waz;
	} else {
		state.info = &default_info;
		state.qso.continent = NULL;
		state.qso.dxcc = NULL;
		state.qso.itu = 0;
		state.qso.waz = 0;
	}

	recalculate_location();
	recalculate_distance();
}

static void cmd_freq(char *arg)
{
	qso_set_freq(&state.qso, arg);
}

static void cmd_mode(char *arg)
{
	qso_set_mode(&state.qso, arg);
}

static void cmd_rst_s(char *arg)
{
	qso_set_rst(&state.qso, false, arg);
}

static void cmd_rst_r(char *arg)
{
	qso_set_rst(&state.qso, true, arg);
}

static void cmd_name(char *arg)
{
	/* TODO: sanitize the input */

	str_putref(state.qso.name);
	state.qso.name = STR_DUP(arg);
}

static void cmd_qth(char *arg)
{
	/* TODO: sanitize the input */

	str_putref(state.qso.qth);
	state.qso.qth = STR_DUP(arg);
}

static void cmd_grid(char *arg)
{
	qso_set_grid(&state.qso, false, arg);

	recalculate_location();
	recalculate_distance();
}

static void cmd_state(char *arg)
{
	/* TODO: sanitize the input */

	str_putref(state.qso.state);
	state.qso.state = STR_DUP(arg);
}

static void cmd_county(char *arg)
{
	/* TODO: sanitize the input */

	str_putref(state.qso.county);
	state.qso.county = STR_DUP(arg);
}

static void cmd_continent(char *arg)
{
	qso_set_zone_continent(&state.qso, arg);
}

static void cmd_dxcc(char *arg)
{
	qso_set_zone_dxcc(&state.qso, arg);
}

static void cmd_itu(char *arg)
{
	qso_set_zone_itu(&state.qso, arg);
}

static void cmd_waz(char *arg)
{
	qso_set_zone_waz(&state.qso, arg);
}

static void cmd_comment(char *arg)
{
	/* TODO: sanitize the input */

	str_putref(state.qso.comment);
	state.qso.comment = STR_DUP(arg);
}

static void print_field(WINDOW *win, int row, int col, const char *str,
			size_t width)
{
	ssize_t len;

	wmove(win, row, col);

	if (!str)
		str = "";

	/*
	 * HACK: count number of printable characters
	 *
	 * We get UTF-8 strings passed in, but we need to know how wide it
	 * will print.  We assume that combining characters don't exist, and
	 * just count the number of bytes < 0x80 because that's how many
	 * codepoints there are.
	 */
	len = utf8_is_valid_str(str, strlen(str));
	ASSERT3S(len, >=, 0);

	if (len <= width) {
		size_t i;

		/* replace leading spaces with '_' */
		for (i = 0; (i < width) && (str[i] == ' '); i++)
			wprintw(win, "_");

		wprintw(win, "%s", &str[i]);

		/* padd with '_' */
		for (i = len; i < width; i++)
			wprintw(win, "_");
	} else {
		/* truncate & append '+' */
		wprintw(win, "%.*s+", width - 1, str);
	}
}

static int print_label(WINDOW *win, int row, int col, const char *str)
{
	bool emph;
	int count;
	size_t i;

	if (!str)
		return 0;

	wmove(win, row, col);

	for (i = 0, emph = false, count = 0; str[i]; i++) {
		if (str[i] == '_') {
			emph = true;
			continue;
		}

		waddch(win, str[i] | COLOR_PAIR(CMD_HINT_COLOR) |
		       (emph ? A_UNDERLINE : A_NORMAL));

		count++;
		emph = false;
	}

	return count;
}

static void setup_windows(void)
{
	size_t i;

	entry_win = newwin(ENTRY_WIN_HEIGHT, ENTRY_WIN_WIDTH, ENTRY_WIN_ROW, 0);
	box(entry_win, 0, 0);

	for (i = 0; i < ARRAY_LEN(entry_fields); i++) {
		const struct entry_field *entry = &entry_fields[i];

		print_label(entry_win, entry->row + 1, entry->col_label + 1,
			    entry->label);
	}

	/* TODO: print '-' to separate the times */

	call_win = newwin(CALL_WIN_HEIGHT, CALL_WIN_WIDTH, CALL_WIN_ROW, 0);
	box(call_win, 0, 0);
	mvwprintw(call_win, 1, 1, "This window will show info about callsign "
		  "being worked");

	recent_win = newwin(RECENT_WIN_HEIGHT, RECENT_WIN_WIDTH, RECENT_WIN_ROW, 0);
	box(recent_win, 0, 0);
	mvwprintw(recent_win, 1, 1, "This window will list the last N contacts");

	cmd_win = newwin(CMD_WIN_HEIGHT, CMD_WIN_WIDTH, CMD_WIN_ROW, 0);
	box(cmd_win, 0, 0);

	mvwprintw(cmd_win, 1, 1, "> ");
}

static void refresh_windows(void)
{
	wrefresh(entry_win);
	wrefresh(call_win);
	wrefresh(recent_win);
	wrefresh(cmd_win);
}

static void refresh_fields(enum refresh refresh)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(entry_fields); i++) {
		const struct entry_field *entry = &entry_fields[i];
		uintptr_t ptr;

		if (!(entry->refresh & refresh))
			continue;

		ptr = ((uintptr_t) &state) + entry->offset;

		switch (entry->type) {
			case VT_STR:
				print_field(entry_win, entry->row + 1,
					    entry->col_value + 1,
					    str_cstr(*(struct str **) ptr),
					    entry->len);
				break;
			case VT_INT: {
				uint64_t v = *(uint64_t *) ptr;
				char buf[22];

				/* default to the empty string */
				buf[0] = '\0';

				switch (entry->refresh) {
					case REF_DE_POWER: {
						const char *unit;
						int width;

						if (!v)
							break; /* no power */
						else if (v < 1000) {
							unit = "mW";
							width = 3;
						} else {
							unit = "W";
							width = 4;

							/* round up */
							if ((v % 1000) >= 500)
								v += 1000;

							v /= 1000;
						}

						snprintf(buf, sizeof(buf),
							 "%*"PRIu64"%s",
							 width, v, unit);
						break;
					}
					case REF_FREQ: {
						uint64_t mhz = v / 1000000;
						uint64_t hz = v % 1000000;

						if (!v)
							break; /* no freq */

						/*
						 * MHz can be used as-is,
						 * but we want to omit as
						 * many zeros after the
						 * decimal point as
						 * possible.  Therefore, we
						 * tweak the Hz value and
						 * rely to the following
						 * snprintf to make them
						 * reasonable looking.
						 */

						while (hz && ((hz % 10) == 0))
							hz /= 10;

						snprintf(buf, sizeof(buf),
							 "%"PRIu64".%"PRIu64,
							 mhz, hz);
						break;
					}
					case REF_ZONE_ITU:
					case REF_ZONE_WAZ:
						if (v)
							snprintf(buf, sizeof(buf),
								 "%02"PRIu64, v);
						break;
					case REF_DE:
					case REF_DE_GRID:
					case REF_TIMES:
					case REF_CALL:
					case REF_MODE:
					case REF_RSTS:
					case REF_RSTR:
					case REF_NAME:
					case REF_QTH:
					case REF_GRID:
					case REF_INFO:
					case REF_GEO:
					case REF_STATE:
					case REF_COUNTY:
					case REF_ZONE_CONT:
					case REF_ZONE_DXCC:
					case REF_COMMENT:
						panic("impossible");
				}

				print_field(entry_win, entry->row + 1,
					    entry->col_value + 1,
					    buf, entry->len);
				break;
			}
			default:
				panic("unhandled entry value type: %d", entry->type);
		}
	}
}

static void parse_command(char *buf)
{
	char *cmdname;
	size_t i;

	/* TODO: strip leading whitespace from buf - the command name */

	cmdname = strsep(&buf, " \t");

	/* TODO: strip leading and trailing whitespace from buf - the arg */

	for (i = 0; i < ARRAY_LEN(commands); i++) {
		const struct command *cmd = &commands[i];

		if (strcasecmp(cmdname, cmd->cmd))
			continue;

		if (cmd->arg_required != (buf && strlen(buf)))
			continue; /* FIXME: report missing arg */

		cmd->fxn(buf);

		refresh_fields(cmd->refresh);
		return;
	}

	/* FIXME: report command not found */
}

static void usage(const char *prog)
{
	fprintf(stderr, "HLog version %s (libjeffpc %s)\n\n", version_string,
		jeffpc_version);

	fprintf(stderr, "Usage: %s <path to data dir>\n", prog);
	exit(1);
}

static int load_countries(const char *dir)
{
	char path[FILENAME_MAX];
	int ret;

	snprintf(path, sizeof(path), "%s/cty.csv", dir);

	printf("Loading countries data...\n");
	ret = country_load(path);
	printf("%s.\n", ret ? "error" : "done");

	return ret;
}

static int load_config(const char *dir)
{
	char path[FILENAME_MAX];
	struct val *parsed;
	const char *raw;
	size_t len;
	int ret;

	snprintf(path, sizeof(path), "%s/config.scm", dir);

	raw = read_file_len(path, &len);
	if (IS_ERR(raw))
		return PTR_ERR(raw);

	parsed = sexpr_parse(raw, len);
	if (IS_ERR(parsed)) {
		ret = PTR_ERR(parsed);
		goto err;
	}

	/* extract the values (NULL strings and 0 ints are ok) */
	state.qso.me.call = sexpr_alist_lookup_str(parsed, "call");
	state.qso.me.grid = sexpr_alist_lookup_str(parsed, "grid");
	state.qso.me.power = sexpr_alist_lookup_int(parsed, "power", NULL);

	val_putref(parsed);

	ret = 0;

err:
	free((void *) raw);

	return ret;
}

int main(int argc, char **argv)
{
	int ret;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	if (argc != 2)
		usage(argv[0]);

	data_dir = argv[1];
	data_dir_fd = xopen(data_dir, O_RDONLY, 0);
	if (data_dir_fd < 0) {
		fprintf(stderr, "Error: failed to open data directory: %s\n",
			xstrerror(data_dir_fd));
		goto err;
	}

	/* TODO: load bands.scm */

	ret = load_countries(data_dir);
	if (ret) {
		fprintf(stderr, "Error: failed to load country data: %s\n",
			xstrerror(ret));
		goto err_close;
	}

	qso_reset(&state.qso, true);

	ret = load_config(data_dir);
	if (ret) {
		fprintf(stderr, "Error: failed to load config: %s\n",
			xstrerror(ret));
		goto err_close;
	}

	initscr();
	start_color();

	init_pair(CMD_HINT_COLOR, COLOR_YELLOW, COLOR_BLACK);

	setup_windows();
	refresh_fields(~0u);
	refresh_windows();

	for (;;) {
		char buf[CMD_WIN_WIDTH - 2 /* border */ - 2 /* prompt */ - 1];

		/* TODO: add line editing support */

		if (wgetnstr(cmd_win, buf, sizeof(buf)) != ERR)
			parse_command(buf);

		/* TODO: clear previous command */

		wmove(cmd_win, 1, 3);
		refresh_windows();
	}

	delwin(entry_win);
	delwin(call_win);
	delwin(recent_win);

	endwin();

	xclose(data_dir_fd);

	return 0;

err_close:
	xclose(data_dir_fd);

err:
	return 1;
}
